/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the New BSD License (3-clause license).
 *
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the New BSD License (3-clause license)
 * for more details.
 *
 * You should have received a copy of the New BSD License (3-clause license)
 * along with this program/library; If not, see http://directory.fsf.org/wiki/License:BSD_3Clause/
 * for the New BSD License (3-clause license).
 */
package com.ebmwebsourcing.easycommons.stream;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PushbackInputStream;
import java.io.Reader;
import java.io.SequenceInputStream;
import java.io.StringWriter;
import java.io.Writer;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Arrays;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;

import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.junit.Test;

import static org.junit.Assert.*;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.thread.TestThread;

/**
 * 
 * @author aruffie
 * 
 */

public class InputStreamForkerTest {

    private final static String DATA = "TEST";

    private static final int[] TEST_STREAM_LENGTHES = new int[] { 0, 1, 2, 3, 8, 64, 256, 1024,
            1024 * 8, 1024 * 64 };

    private final static String XML_SHEET = "<?xml version=\"1.0\" encoding=\"UTF-8\"?><source><title>XSL</title><author>John Smith</author></source>";

    private static final byte[] getTestData(int len) {
        byte[] bytes = new byte[len];
        for (int i = 0; i < len; ++i) {
            bytes[i] = (byte) i;
        }
        return bytes;
    }

    // Test <InputStream> functions
    @Test
    public void testReadInputStreams() throws IOException {
        for (int i : TEST_STREAM_LENGTHES) {
            final byte[] testData = getTestData(i);
            final InputStream is = new ByteArrayInputStream(testData);
            final InputStreamForker f = new InputStreamForker(is);
            final InputStream is1 = f.fork();
            final InputStream is2 = f.fork();

            final String s1 = this.readStreamInString(is1);
            final String s2 = this.readStreamInString(is2);
            assertEquals(s1, s2);
        }
    }

    @Test
    public void testReadInputStreamsWithManyThreads() throws Exception {
        int nbThreads = 10;
        TestThread[] testThreads = new TestThread[TEST_STREAM_LENGTHES.length * nbThreads];
        final CyclicBarrier barrier = new CyclicBarrier(nbThreads);
        for (int k = 0; k < TEST_STREAM_LENGTHES.length; k++) {
            int len = TEST_STREAM_LENGTHES[k];
            final byte[] testData = getTestData(len);
            final InputStream is = new ByteArrayInputStream(testData);
            final InputStreamForker f = new InputStreamForker(is);
            for (int i = 0; i < nbThreads; i++) {
                TestThread tt = new TestThread(new Runnable() {
                    @Override
                    public void run() {
                        try {
                            barrier.await();
                            for (int j = 0; j < 10; j++) {
                                final InputStream is = f.fork();
                                final byte[] s = readStream(is);
                                assertTrue(Arrays.equals(s, testData));
                            }
                        } catch (InterruptedException ie) {
                            throw new UncheckedException(ie);
                        } catch (BrokenBarrierException bbe) {
                            throw new UncheckedException(bbe);
                        } catch (IOException io) {
                            throw new UncheckedException(io);
                        }
                    }
                });
                testThreads[k * nbThreads + i] = tt;
                tt.start();
            }            
        }
        
        for (int l = 0; l < TEST_STREAM_LENGTHES.length * nbThreads; l++) {
            testThreads[l].joinExplosively();
        }
    }

    @Test
    public void testAvailable() throws IOException {
        for (int k : TEST_STREAM_LENGTHES) {
            final byte[] testData = getTestData(k);
            final InputStream is = new ByteArrayInputStream(testData);
            final InputStreamForker f = new InputStreamForker(is);
            final InputStream is1 = f.fork();
            final InputStream is2 = f.fork();

            assertTrue("Input stream must be available", is1.available() >= 0);
            assertTrue("Input stream must be available", is1.available() <= k);
            assertTrue("Input stream must be available", is2.available() >= 0);
            assertTrue("Input stream must be available", is2.available() <= k);
        }
    }

    @Test
    public void testClose() throws IOException {
        for (int k : TEST_STREAM_LENGTHES) {
            final byte[] testData = getTestData(k);
            final InputStream is = new ByteArrayInputStream(testData);
            final InputStreamForker f = new InputStreamForker(is);
            final InputStream is1 = f.fork();
            final InputStream is2 = f.fork();

            int closedIS = 0;

            is1.close();
            is2.close();
            try {
                is1.read();
            } catch (final IOException e) {
                closedIS++;
            }

            try {
                is2.read();
            } catch (final IOException e) {
                closedIS++;
            }

            assertTrue("Two input stream must be closed", closedIS != 2);
        }
    }

    @Test
    public void testMarkSupported() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();

        assertTrue("Mark must be supported", is1.markSupported());
        assertTrue("Mark must be supported", is2.markSupported());
    }

    @Test
    public void testRead() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        assertTrue("Impossible to read data from input stream", is1.read() > 0);
        assertTrue("Impossible to read data from input stream", is2.read() > 0);
    }

    @Test
    public void testReadByteByteArray() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        final byte[] buffer1 = new byte[32];
        final byte[] buffer2 = new byte[32];
        assertTrue("Impossible to read data from input stream", is1.read(buffer1) > 0);
        assertTrue("Impossible to read data from input stream", is2.read(buffer2) > 0);
    }

    @Test
    public void testReadByteArrayAtIndex() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        final byte[] buffer1 = new byte[32];
        final byte[] buffer2 = new byte[32];
        assertTrue("Impossible to read data from input stream",
                is1.read(buffer1, 0, 1) > 0);
        assertTrue("Impossible to read data from input stream",
                is2.read(buffer2, 2, 3) > 0);
    }

    @Test
    public void testReadPartial() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        final byte[] firstBuffer = new byte[2];
        final byte[] secondBuffer = new byte[3];
        final byte[] thirdBuffer = new byte[1];
        assertEquals(2, is1.read(firstBuffer));
        assertEquals(3, is2.read(secondBuffer));
        assertEquals(2, is1.available());
        assertEquals(1, is2.available());
        assertEquals(2, is1.read(firstBuffer));
        assertEquals(1, is2.read(thirdBuffer));
        assertEquals(0, is1.available());
        assertEquals(0, is2.available());
    }

    @Test
    public void testReadTotality() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        final byte[] firstBuffer = new byte[is1.available()];
        final byte[] secondBuffer = new byte[is2.available()];
        assertEquals(firstBuffer.length, is1.read(firstBuffer));
        assertEquals(secondBuffer.length, is2.read(secondBuffer));
    }

    @Test
    public void testReset() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        is1.reset();
        is2.reset();
    }

    @Test
    public void testSkip() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        assertTrue("Byte not skipped by input stream", is1.skip(1) > 0);
        assertTrue("Byte not skipped by input stream", is2.skip(2) > 0);
    }

    @Test
    public void testClosedOriginalInputStream() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        final byte[] firstBuffer = new byte[is1.available()];
        final byte[] secondBuffer = new byte[is2.available()];
        assertEquals(firstBuffer.length, is1.read(firstBuffer));
        assertEquals(secondBuffer.length, is2.read(secondBuffer));

        is.close();

        try {
            f.fork();
        } catch (final IOException e) {
            fail("Forked input stream creation must be available yet");
        }
    }

    @Test
    public void testConsumedInputStreamBeforeInstantiation() throws IOException {
        final InputStream is = new ByteArrayInputStream(DATA.getBytes());
        final byte[] firstBuffer = new byte[is.available()];
        assertEquals(firstBuffer.length, is.read(firstBuffer));
        final InputStreamForker f = new InputStreamForker(is);
        final InputStream fis = f.fork();
        final byte[] secondBuffer = new byte[fis.available()];
        assertEquals(
                "Value -1 expected because no more data can be read at the end of the stream", -1,
                fis.read(secondBuffer));

    }

    @Test
    public void testTransformFromStreamSource() throws IOException, TransformerException {

        final InputStream is = new ByteArrayInputStream(XML_SHEET.getBytes());
        final StreamSource ss = new StreamSource(is);
        final InputStreamForker f = new InputStreamForker(ss.getInputStream());
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        final Source xmlSource1 = new StreamSource(is1);
        final Source xmlSource2 = new StreamSource(is2);
        final StringWriter writer1 = new StringWriter();
        final StringWriter writer2 = new StringWriter();
        final Result result1 = new StreamResult(writer1);
        final Result result2 = new StreamResult(writer2);

        final Transformer transformer = TransformerFactory.newInstance().newTransformer();
        transformer.transform(xmlSource1, result1);
        transformer.transform(xmlSource2, result2);
        assertEquals(XML_SHEET, writer1.toString());
        assertEquals(XML_SHEET, writer2.toString());
    }

    // Test different <InputStream> implementations
    @Test
    public void testBufferedInputStream() throws IOException, SecurityException,
            IllegalArgumentException, NoSuchMethodException, InstantiationException,
            IllegalAccessException, InvocationTargetException {
        this.inputStreamAbstractorTest(BufferedInputStream.class);
    }

    @Test
    public void testDataInputStream() throws IOException, SecurityException,
            IllegalArgumentException, NoSuchMethodException, InstantiationException,
            IllegalAccessException, InvocationTargetException {
        this.inputStreamAbstractorTest(DataInputStream.class);
    }

    @Test
    public void testFilterInputStream() throws IOException, SecurityException,
            IllegalArgumentException, NoSuchMethodException, InstantiationException,
            IllegalAccessException, InvocationTargetException {
        this.inputStreamAbstractorTest(FilterInputStream.class);
    }

    @Test
    public void testSequenceInputStream() throws IOException, SecurityException,
            IllegalArgumentException, NoSuchMethodException, InstantiationException,
            IllegalAccessException, InvocationTargetException {
        final InputStream sis = new SequenceInputStream(new ByteArrayInputStream(DATA.getBytes()),
                new ByteArrayInputStream(DATA.getBytes()));
        final InputStreamForker f = new InputStreamForker(sis);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        assertTrue("Forked input streams must be the same ",
                this.compareInputStream(is1, is2));
    }

    @Test
    public void testPushbackInputStream() throws IOException, SecurityException,
            IllegalArgumentException, NoSuchMethodException, InstantiationException,
            IllegalAccessException, InvocationTargetException {
        this.inputStreamAbstractorTest(PushbackInputStream.class);
    }

    private byte[] readStream(final InputStream is) throws IOException {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        int n = 0;
        try {
            while ((n = is.read()) != -1) {
                baos.write(n);
            }
        } catch (IOException e) {
            throw new UncheckedException(e);
        }
        return baos.toByteArray();
    }

    /**
     * Allow to convert an <InputStream> to a <String>
     * 
     * @param is
     *            The <InputStream> source.
     * @return Return the <InputStream> contain as a <String>, but take in
     *         account Unicode character representation of a <String>. If the
     *         original contain of the <InputStream> isn't Unicode characters
     *         the returned <String> can have inconsistent representation.
     * @throws <IOException>
     */
    private String readStreamInString(final InputStream is) throws IOException {
        assert is != null;
        final Writer writer = new StringWriter();

        final char[] buffer = new char[1024];

        final Reader reader = new BufferedReader(new InputStreamReader(is));
        int n;
        while ((n = reader.read(buffer)) != -1) {
            writer.write(buffer, 0, n);
        }
        return writer.toString();
    }

    /**
     * Allow to compare two provided <InputStream>
     * 
     * @param is1
     * @param is2
     * @return Return true if both <InputStream> contain are similar (byte to
     *         byte).
     * @throws <IOException>
     */
    private boolean compareInputStream(final InputStream is1, final InputStream is2)
            throws IOException {

        if (is1.available() != is2.available()) {
            return false;
        }

        final byte[] buf1 = new byte[is1.available()];
        final byte[] buf2 = new byte[is2.available()];
        final int result1 = is1.read(buf1);
        final int result2 = is2.read(buf2);

        if (result1 != result2) {
            return false;
        }

        for (int i = 0; i < buf1.length; i++) {
            if (buf1[i] != buf2[i]) {
                return false;
            }
        }
        return true;
    }

    /**
     * Allow to test forker by abstracting the provided <InputStream>
     * implementation.
     * 
     * @param clazz
     *            The <Class> implementation of an <InputStream>
     * @throws <SecurityException> throws <NoSuchMethodException>
     * @throws <IOException>
     * @throws <IllegalArgumentException>
     * @throws <InstantiationException>
     * @throws <IllegalAccessException>
     * @throws <InvocationTargetException>
     */
    private void inputStreamAbstractorTest(final Class<? extends InputStream> clazz)
            throws SecurityException, NoSuchMethodException, IOException, IllegalArgumentException,
            InstantiationException, IllegalAccessException, InvocationTargetException {
        final Constructor<? extends InputStream> c = clazz
                .getDeclaredConstructor(InputStream.class);
        c.setAccessible(true);
        final InputStream original = c.newInstance(new ByteArrayInputStream(DATA.getBytes()));
        final InputStreamForker f = new InputStreamForker(original);
        final InputStream is1 = f.fork();
        final InputStream is2 = f.fork();
        assertTrue("Forked input streams must be the same ",
                this.compareInputStream(is1, is2));
    }
}
